open Sasacore
(* open SasArg *)
 
let (get_action_value : (string * Data.v) list -> string -> string -> bool) =
  fun sl pid a ->
    let vn = Printf.sprintf "%s_%s" pid (StringOf.action a) in
    match List.assoc_opt vn sl with
    | Some (Data.B b) -> b
    | None 
    | Some _ ->
      Printf.printf "Error: can not find %s in [%s]\n" vn
        (String.concat "," (fst (List.split sl)));
      flush stdout;
      assert false

open Sasacore
open Process
let (from_sasa_env : 'v SimuState.t -> RdbgPlugin.sl) =
  fun st ->
    List.fold_left
      (fun acc p ->
         let state = Env.get st.config p.pid in
         let sl = SasaState.to_rdbg_subst p.pid state in
         acc@sl
      )
      []
      st.network

let (get_sl_out: bool -> 'v Process.t list -> bool list list -> RdbgPlugin.sl) =
  fun enab pl ll ->
    let str = if enab then "Enab_" else "" in
    List.flatten (
      List.map2 (fun p enab_l ->
          List.map2 (fun a enab ->
              Printf.sprintf "%s%s_%s" str p.pid (StringOf.action a), Data.B enab)
            p.actions enab_l)
        pl ll
    )

module StringMap = Map.Make(String)
let (compute_potentiel: 'v SimuState.t -> RdbgPlugin.sl) =
  fun st ->
  match Register.get_potential () with
  | None -> []
  | Some user_pf ->
    let pidl = List.map (fun p -> p.Process.pid) st.network in
    let p = user_pf pidl (SimuState.neigbors_of_pid st) in
    [("potential", Data.F p)]
    
let (compute_legitimate: bool -> 'v SimuState.t -> bool) =
  fun silent st ->
  silent ||
  match Register.get_legitimate () with
  | None -> silent 
  | Some f ->
    let pidl = List.map (fun p -> p.Process.pid) st.network in
    f pidl (SimuState.neigbors_of_pid st)

open SimuState
let (make_do: string array -> 'v SimuState.t -> RdbgPlugin.t) =
  fun argv st ->
  let pl = st.network in
  let prog_id = Printf.sprintf "%s (with sasa Version %s)"
      (String.concat " " (Array.to_list argv)) SasaVersion.str
  in
  let vntl_i =
    List.map (fun (vn,vt) -> vn, Data.type_of_string vt)
      (Sasacore.SimuState.get_inputs_rif_decl st.sasarg pl)
  in
  let vntl_o =
    List.map (fun (vn,vt) -> vn, Data.type_of_string vt)
      (Sasacore.SimuState.get_outputs_rif_decl st.sasarg pl)
  in
  let vntl_o =
    if Register.get_potential () = None then vntl_o else
      ("potential", Data.Real)::vntl_o in
  let vntl_o = ("silent", Data.Bool)::("legitimate", Data.Bool)::vntl_o in
  let pre_enable_processes_opt = ref None in
  let sasa_config = ref st.config in
  let reset () =
    pre_enable_processes_opt := None;
    sasa_config :=
      (Sasacore.SimuState.make false argv).config (* to take seed changes into account *)
  in
  (*  Do the same job as SasaSimuState.simustep *)
  let (step_custom: RdbgPlugin.sl -> RdbgPlugin.sl) =
    fun sl_in ->
      let st = Sasacore.SimuState.update_config !sasa_config st in
      match !pre_enable_processes_opt with
      | None  -> ( (* the first step *)
          (* 1: Get enable processes *)
          let pnall, enab_ll = Sasacore.SimuState.get_enable_processes st in
          let sasa_nenv = from_sasa_env st in 
          let pot_sl = compute_potentiel st in
          let silent = List.for_all (fun b -> not b) (List.flatten enab_ll) in
          let legit = compute_legitimate silent st in
          pre_enable_processes_opt := Some(pnall, enab_ll);
          ("silent", Data.B silent)::("legitimate", Data.B legit)::pot_sl
          @ sasa_nenv @ (get_sl_out true pl enab_ll)
        )
      | Some (pre_pnall, pre_enab_ll) ->
        (* let was_silent = List.for_all (fun b -> not b) (List.flatten pre_enab_ll) in *)
        (* if was_silent then failwith "Silent"; *)
        (* 2: read the actions from the outside process, i.e., from sl_in *)
        let _, pnal = Daemon.f st.sasarg.dummy_input
            (st.sasarg.verbose > 0) st.sasarg.daemon st.network
            SimuState.neigbors_of_pid st pre_pnall pre_enab_ll
            (get_action_value sl_in) Step.f
        in
        (* 3: Do the steps *)
        let st = Sasacore.Step.f pnal st in
        let nst = update_config st.config st in
        let sasa_nenv = from_sasa_env nst in
        (* 1': Get enable processes *)
        let pnall, enab_ll = Sasacore.SimuState.get_enable_processes nst in
        let pot_sl = compute_potentiel nst in
        let silent = List.for_all (fun b -> not b) (List.flatten enab_ll) in
        let legit = compute_legitimate silent nst in
        pre_enable_processes_opt := Some(pnall, enab_ll);
        sasa_config := st.config;    
        ("silent", Data.B silent)::("legitimate", Data.B legit)::pot_sl @
        sasa_nenv @ (get_sl_out true pl enab_ll) 
  in
  let (step_internal_daemon: RdbgPlugin.sl -> RdbgPlugin.sl) =
    fun sl_in ->
      (* in this mode, sasa does not play first *)
      let st = update_config !sasa_config st in
      (* 1: Get enable processes *)
      let pnall, enab_ll = Sasacore.SimuState.get_enable_processes st in
      let pot_sl = compute_potentiel st in
      let silent = List.for_all (fun b -> not b) (List.flatten enab_ll) in
      let legit = compute_legitimate silent st in
      if silent then (
        let enab = get_sl_out true pl enab_ll in
        let trigger = (* set all trig var to false *)
          List.map (fun (n,v) -> String.sub n 5 ((String.length n) -5),v) enab 
        in        
        ("silent", Data.B silent)::("legitimate", Data.B legit)::pot_sl@
        (from_sasa_env st) @ enab @ trigger
      )
      else
        (* 2: read the actions from the outside process, i.e., from sl_in *)
        let activate_val, pnal =
          Daemon.f st.sasarg.dummy_input
            (st.sasarg.verbose > 0) st.sasarg.daemon st.network
            SimuState.neigbors_of_pid st pnall enab_ll
            (get_action_value sl_in) Step.f
        in
        (* 3: Do the steps *)
        let st = Sasacore.Step.f pnal st in
        sasa_config := st.config;
        ("silent", Data.B silent)::("legitimate", Data.B legit)::pot_sl @
        (from_sasa_env st) @ (get_sl_out true pl enab_ll) @
        (get_sl_out false pl activate_val)
  in
  let step =
    if st.sasarg.daemon = DaemonType.Custom then
      step_custom
    else
      step_internal_daemon
  in
  let ss_table = Hashtbl.create 10 in
  let step_dbg sl_in ctx cont =
    let sl_out = step sl_in in 
    { ctx with
      (* RdbgEvent.nb = 0; *)
      (* RdbgEvent.step = 0; (* we are actually in the middle of the first step! *) *)
      RdbgEvent.depth = ctx.RdbgEvent.depth + 1;
      RdbgEvent.kind = RdbgEvent.Exit;
      RdbgEvent.lang = "sasa";
      RdbgEvent.sinfo = None;
      RdbgEvent.name = "sasa";
      RdbgEvent.inputs = vntl_i;
      RdbgEvent.outputs = vntl_o;
      RdbgEvent.locals = [];
      RdbgEvent.data = sl_in@sl_out;
      RdbgEvent.next = (
        fun () ->
          let ctx = { ctx with RdbgEvent.nb = 1 + ctx.RdbgEvent.nb } in
          cont sl_out ctx );
    }
  in
  let (mems_in  : Data.subst list) = [] in
  let (mems_out : Data.subst list) = [] in
  {
    id = prog_id;
    inputs = vntl_i;
    outputs= vntl_o;
    reset=(fun () -> reset());
    kill=(fun _ -> flush stdout; flush stderr);
    init_inputs=mems_in;
    init_outputs=mems_out;
    step=step;     
    step_dbg = step_dbg;
    save_state = (fun i ->
        let prgs = Random.get_state () in
        (* Printf.eprintf "Save state %i from sasa\n%!" i; *)
        Hashtbl.replace ss_table i
          (prgs, !sasa_config, !pre_enable_processes_opt)
      );
    restore_state = (fun i ->
        match Hashtbl.find_opt ss_table i with
        | Some (prgs, e, pepo) ->
          Random.set_state prgs;
          (* Printf.eprintf "Restore state %i from sasa\n%!" i; *)
          sasa_config := e; pre_enable_processes_opt := pepo
        | None ->
          Printf.eprintf "Cannot restore state %i from sasa\n" i;
          flush stderr
      );
  }


let (make: string array -> RdbgPlugin.t) =
  fun argv ->
    try
      make_do argv (Sasacore.SimuState.make false argv)
    with
    | Dynlink.Error e ->
      Printf.printf "Error (SasaRun.make): %s\n" (Dynlink.error_message e);
      flush stdout;
      exit 2
    | Failure msg ->
      Printf.printf "Error (SasaRun.make): %s\n" msg;
      flush stdout;
      exit 2

    | Assert_failure (file, line, col)  -> 
      prerr_string (
        "\nError: oops, sasa internal error\n\tFile \""^ file ^
        "\", line " ^ (string_of_int line) ^ ", column " ^
        (string_of_int col) ^ "\n") ; 
      flush stderr;
      exit  2
