(* Time-stamp: <modified the 11/10/2021 (at 16:38) by Erwan Jahier> *)

open Register

type 's node = { st : 's ; d : int ; cost : int ; cpt : int }
type point = value Array.t

let debug = false

          (* type succ_heuristic = (* Various heuristic to choose the neighbor *) *)
          (*   | OneDim (* only move one value at a time *) *)
          (*   | RanDim (* move a value or not at random *) *)
(*   | AllDim (* move all values *) *)
      
(* let succ_heuristic_nb = 3 *)
(* let int_of_succ_heuristic = function OneDim -> 0 | RanDim -> 1 | AllDim -> 2 *)
  (* let succ_heuristic_of_int = function 0 -> OneDim | 1 -> RanDim | 2 -> AllDim *)
(*                               | _  -> assert false *)

let deltaf=1.0 (* something else ?? *)
let mutate_value = function
  | F (minf,f,maxf) ->
    let nf =
    if minf > f -. deltaf then
      minf+.(Random.float (2.0 *. deltaf))
    else if maxf < f +. deltaf then
      maxf -. (Random.float (2.0 *. deltaf))
    else
      f+.(Random.float (2.0 *. deltaf)) -. deltaf
    in
    F(minf, nf, maxf)

  | I (mini,i,maxi) ->
    let ni = if i=mini then i+1 else if i=maxi then i-1 else
      if Random.bool () then i+1 else i-1
    in
    I (mini,ni,maxi)
  | B b -> B (not b)
                                
                                
let one_dim_succ n =
  let p = Array.copy n in
  let j = Random.int (Array.length p) in
  p.(j) <- mutate_value p.(j);
  p


let ran_dim_succ n =
  let p = Array.copy n in
  for j=0 to Array.length p - 1 do
    if Random.bool () then p.(j) <- mutate_value p.(j)
  done;
  p

let all_dim_succ n =
  let p = Array.copy n in
  for j=0 to Array.length p - 1 do
    p.(j) <- mutate_value p.(j)
  done;
  p
    
      
(* let tf = float_of_int  *)

let (choose : int -> point -> (point -> point) -> point list) =
  fun n p heuristic ->
  (*choose n successors of p using heuristic *)  
  assert (n>=0);
  let rec f acc i = if i <= 0 then acc else f ((heuristic p)::acc)  (i-1) in
  f [] n

(*****************************************************************************)
(* XXX a ranger ailleurs !!! *)

open Process
    
let (point_to_ss : point -> 'v SimuState.t -> 'v SimuState.t) =
  fun point ss ->
  let (state_to_values, values_to_state :
       ('v -> Register.value list) * (Register.value list -> 'v -> 'v )) =
    match Register.get_init_search_utils () with
    | None -> assert false
    | Some (f, g) -> f, g
  in
  let state_size =
    assert (ss.network <> []);
    let p0 = List.hd ss.network in
    let s0 = state_to_values p0.init in
    List.length s0
  in
  let make_value l i j =
    if debug then
      Printf.printf "make_value  i=%d j=%d size=%dx%d \n%!"
        i j (Array.length point) state_size;
    let rec f l i j =
      if i=0 then l else f (point.(i+j-1)::l) (i-1) j
    in
    f l i j
  in
  let new_config, _ = 
    List.fold_left 
    (fun (e,j) p ->
       let value = make_value [] state_size (j*state_size) in
       let st = values_to_state value (Env.get e p.pid) in
       let e = Env.set e p.pid st in
       e, j+1
    )
    (ss.config, 0)
    ss.network
  in
  if debug then Printf.printf "point_to_ss  ok\n%!";
  { ss with config = new_config }

    
let (ss_to_point : 'v SimuState.t -> point) =
  fun ss ->
  let (state_to_values : ('v -> Register.value list) ) =
    match Register.get_init_search_utils () with
      | None ->
        failwith ("the Algo.init_search_utils registration field should provide "^
                  "state_to_values functions")
      | Some (f, _) -> f
  in
  let size =
    assert (ss.network <> []);
    let p0 = List.hd ss.network in
    let s0 = state_to_values p0.init in
    (List.length s0) * (List.length ss.network)
  in
  let point = Array.make size (I (0,0,0) (* fake unused value *) ) in
    let i = ref 0 in
    List.iter
    (fun p ->
       let values = state_to_values p.init in
       List.iter (fun value ->
           assert (!i<Array.length point);
           point.(!i) <- value; incr i) values
    )
    ss.network;
  point
(*****************************************************************************)
    
open LocalSearch


module ValueArrayNode = struct
  type t = value array node
  let compare n1 n2 = compare n2.cost n1.cost
end

(* open Functory.Sequential *)
open Functory.Cores
module Q = Psq.Make (Int) (ValueArrayNode)

let value2str = function
    Register.F(_, f, _) -> string_of_float f
  | Register.I(_, i, _) -> string_of_int i
  | Register.B true -> "t"
  | Register.B false -> "f"

let point2str p =
  Array.fold_right (fun value acc -> (value2str value)::acc)  p []
  |> String.concat ";"

(* First Choice Hill Climbing: a  successor is chosen at random (using
   some  heuristics), and  became  the  current state  if  its cost  is
   better.

   The heuristic to choose the succ is chosen at random in an array of
   heuristics.   The probability  of  each heuristic  evolves, but  is
   never null.  *)
let (fchc : out_channel -> ('v SimuState.t -> int) -> 'v SimuState.t -> int
        -> 'v SimuState.t) =
  fun log run ss_init dmax ->
  let cpt = ref 0 in  
  let cost p = run (point_to_ss p ss_init) in
  let pinit = ss_to_point ss_init in
  let g =
    { 
      init = ({ st = pinit ; d = 0 ; cost = cost pinit ; cpt = 0 }, Q.empty, ());
      succ = (fun n ->
          let m = min 50 (max 1 (dmax / 10)) in
          let pl = (choose m n.st one_dim_succ) @
                   (choose 1 n.st ran_dim_succ) @
                   (choose 1 n.st all_dim_succ)
          in
            let new_cpt, res =
            map_local_fold
              ~f: (fun p -> cost p, p)
              ~fold:(fun (cpt,nl) (c,p) ->
                  Printf.fprintf log "At depth %d, cost=%d\n%!" (n.d+1) c;
                  cpt+1,{ d=n.d+1; cost=c; st=p ; cpt = cpt}::nl
                )
              (!cpt, [])
              pl
            in
            Printf.fprintf log "fchc: cpt=%d->%d cost=[%s]\n%!" !cpt new_cpt
              (String.concat "," (List.map (fun n -> string_of_int n.cost) res));
            cpt:=new_cpt;
            res
      );
      stop = (fun _ _n -> !cpt > dmax);
      is_goal = (fun _n -> true);
      push = (fun tv n ->
          Printf.fprintf log "Pushing a point of cost %d (queue size=%d)\n%!"
            (n.cost) (Q.size tv);
          Q.add n.cpt n tv);
      pop = (fun tv -> match Q.pop tv with None -> None | Some((i,x),t) ->
             Printf.fprintf log "Poping a point of cost %d (simu #%d)\n%!" x.cost i;
             Some(x, t));
      visiting = (fun _ x -> x);
      visited  = (fun _ _ -> false);
      cut  = (fun pn n -> if pn.cost > n.cost then (
          Printf.fprintf log "%d: cut a point at depth %d of cost %d < %d\n%!"
            !cpt n.d n.cost pn.cost;
          true)
          else false             );
    }
  in
  let rec run_more psol more =
      ( match more (Some psol) with
        | LocalSearch.Stopped 
        | LocalSearch.NoMore->
          Printf.printf "The worst initial configuration, which costs %d, is " psol.cost;
          point_to_ss psol.st ss_init
        | LocalSearch.Sol (nsol, more) ->
          
            if nsol.cost > psol.cost then (
              Printf.printf "Hey, I've found a conf of cost %d! (simu #%d, depth %d)\n%!"
                nsol.cost nsol.cpt nsol.d;
              run_more nsol more
            ) else (
              run_more psol more            
            )
      )
  
  in
  
  match LocalSearch.run g None with
  | LocalSearch.Stopped -> assert false (* SNO *)
  | LocalSearch.NoMore-> assert false (* SNO *)
  | LocalSearch.Sol (sol, more) -> run_more sol more
