open Sasacore

let (print_step : out_channel -> 'v SimuState.t -> int -> int -> string -> string -> SasArg.t ->
      'v Env.t -> 'v Process.t list -> string -> bool list list -> unit) =
  fun log st n i legitimate pot args e pl activate_val enab_ll ->
  let enable_val =
    String.concat " " (List.map (fun b -> if b then "t" else "f")
                         (List.flatten enab_ll))
  in
  if st.sasarg.init_search_max_trials <> None then (
    (* Printf.fprintf log "\n#step %s\n%!" (string_of_int (n-i)); *)
    (* Printf.fprintf log "%s %s %s %s\n%!" (StringOf.env_rif e pl) enable_val legitimate pot; *)
  ) else  
  if args.no_data_file then (
    Printf.printf "\n#step %s\n%!" (string_of_int (n-i))
  ) else (
    if args.daemon = DaemonType.Custom then (
      (* in custom mode, to be able to talk with lurette, this should not be 
           printed on stdout
       *)
      if not args.rif then (
        Printf.eprintf "\n#step %s\n" (string_of_int (n-i)) ;
        Printf.eprintf "%s #outs " activate_val; flush stderr
      );
      Printf.printf "%s %s %s %s\n%!" (StringOf.env_rif e pl) enable_val legitimate pot;
    ) else (
      (* rif mode, internal daemons *)
      if args.rif then
        Printf.printf " %s %s %s %s %s\n%!"
          (StringOf.env_rif e pl) enable_val activate_val legitimate pot
      else (
        Printf.printf "\n#step %s\n" (string_of_int (n-i));
        Printf.printf "%s%s %s %s %s %s\n%!"
          (if args.rif then "" else "#outs ")
          (StringOf.env_rif e pl) enable_val activate_val legitimate pot
      );
    );
    flush stderr;
    flush stdout
  )
  
exception Silent of int
exception Legitimate of int 
let moves  = ref 0
let rounds = ref 0
let round_mask = ref []

let (update_round : bool list list -> bool list list -> unit) =
  fun acti_ll enab_ll ->
  if !round_mask = [] then round_mask := enab_ll;
  round_mask := List.map2 (List.map2 (&&)) !round_mask enab_ll;
  round_mask := List.map2 (List.map2 (fun m a -> m && not a)) !round_mask acti_ll;
  if List.for_all (List.for_all (not)) !round_mask then (
    round_mask := enab_ll;
    incr rounds
  );
  ()

let bool_ll_to_string bll =
  String.concat " " 
    (List.map (fun b -> if b then "t" else "f") (List.flatten bll))

open Sasacore.SimuState

let legitimate st =
  match Register.get_legitimate () with
  | None -> false
  | Some ulf ->
    let pidl = List.map (fun p -> p.Process.pid) st.network in
    ulf pidl  (SimuState.neigbors_of_pid st)

module StringMap = Map.Make(String)

let inject_fault ff st =
  let update_nodes e p =
    let nl = StringMap.find p.Process.pid st.neighbors in
    let pid = p.Process.pid in
    let v = Env.get e pid in
    let v = ff (List.length nl) pid v in
    Env.set e pid v
  in
  let e = List.fold_left update_nodes st.config st.network in
  update_config e st
   
let plur i = if i>1 then "s" else ""

let (compute_potentiel: 'v SimuState.t -> string) =
  fun st ->
  match Register.get_potential () with
  | None -> ""
  | Some user_pf ->
    let pidl = List.map (fun p -> p.Process.pid) st.network in
    let p = user_pf pidl (SimuState.neigbors_of_pid st) in
    string_of_float p

       
let (simustep: out_channel -> int -> int -> string -> 'v SimuState.t -> 'v SimuState.t * string) =
  fun log n i activate_val st ->
  (* 1: Get enable processes *)
  let verb = !Register.verbose_level > 0 in
  if verb then Printf.fprintf log "==> SasaSimuState.simustep :1: Get enable processes\n%!";
  let all, enab_ll = Sasacore.SimuState.get_enable_processes st in
  let pot = compute_potentiel st in
  let pl = st.network in
  let leg = legitimate st in
  let st, all, enab_ll =
    if
      (* not (args.rif) && *)
        List.for_all (fun b -> not b) (List.flatten enab_ll)
    then (
      match Register.get_fault () with
      | None ->
        print_step log st n i "t" pot st.sasarg st.config pl activate_val enab_ll;
        incr rounds;
        raise (Silent (n-i))
      | Some ff ->
        print_step log st n i "t" pot st.sasarg st.config pl activate_val enab_ll;
        let str = if st.sasarg.rif then "#" else "" in
        Printf.fprintf log
          "\n%sThis algo is silent after %i move%s, %i step%s, %i round%s.\n"
          str !moves (plur !moves) (n-i) (plur (n-i)) !rounds (plur !rounds);
        Printf.fprintf log "%s==> Inject a fault\n%!" str;
        let st = inject_fault ff st in
        let all, enab_ll = Sasacore.SimuState.get_enable_processes st in
        st, all, enab_ll
    )
    else if leg then (
      match Register.get_fault () with
      | None ->
        print_step log st n i "t" pot st.sasarg st.config pl activate_val enab_ll;
        raise (Legitimate (n-i))
      | Some ff ->
        print_step log st n i "t" pot st.sasarg st.config pl activate_val enab_ll;
        let str = if st.sasarg.rif then "#" else "#" in
        Printf.fprintf log
          "\n%sThis algo reached a legitimate configuration after %i move%s, %i step%s, %i round%s.\n"
          str !moves (plur !moves) (n-i) (plur (n-i)) !rounds (plur !rounds);
        Printf.fprintf log "%s==> Inject a fault\n%!" str;
        let st = inject_fault ff st in
        let all, enab_ll = Sasacore.SimuState.get_enable_processes st in
        st, all, enab_ll
    )
    else
      st, all, enab_ll
  in
  let leg_str = if leg then "t" else "f" in
  if st.sasarg.daemon = DaemonType.Custom then
    print_step log st n i leg_str pot st.sasarg st.config pl activate_val enab_ll;
  (* 2: read the actions *)
  if verb then Printf.fprintf log "==> SasaSimuState.simustep : 2: read the actions\n%!";
  let get_action_value = RifRead.bool (st.sasarg.verbose > 1) in
  let next_activate_val, pnal = Daemon.f st.sasarg.dummy_input
      (st.sasarg.verbose >= 1) st.sasarg.daemon st.network SimuState.neigbors_of_pid
      st all enab_ll get_action_value Step.f
  in
  List.iter (List.iter (fun b -> if b then incr moves)) next_activate_val;
  update_round next_activate_val enab_ll;
  let next_activate_val = bool_ll_to_string next_activate_val in
  (* 3: Do the steps *)
  if verb then Printf.fprintf log "==> SasaSimuState.simustep : 3: Do the steps\n%!";
  if st.sasarg.daemon <> DaemonType.Custom then
    print_step log st n i leg_str pot st.sasarg st.config pl next_activate_val enab_ll;
  let st = Sasacore.Step.f pnal st in
  st, next_activate_val

let rec (simuloop: out_channel -> int -> int -> string -> 'v SimuState.t -> int) =
  fun log n i activate_val st ->
  let rec loop i activate_val st =
    if !Register.verbose_level > 0 then
      Printf.fprintf log "==> SasaSimuState.simuloop %d/%d \n%!" i n;
    let st, next_activate_val = simustep log n i activate_val st in
    if i > 0 then loop (i-1) next_activate_val st else (
      print_string "#q\n"; flush_all ()
    )
  in
  try (loop i activate_val st); n
  with
  | Silent i ->
    let str = if st.sasarg.rif then "#" else "" in
    Printf.fprintf log
      "\n%sThis algo is silent after %i move%s, %i step%s, %i round%s.\n%!"
      str !moves (plur !moves) i (plur i) !rounds (plur !rounds);
    if st.sasarg.rif then 
      print_string "\nq\n#quit\n%!"
    else
      Printf.fprintf log "\n#quit\n"; 
    flush_all();
    i  
    
  | Legitimate i ->
    let str = if st.sasarg.rif then "#" else "" in
    Printf.fprintf log
      "\n%s%sThis algo reached a legitimate configuration after %i move%s, %i step%s, %i round%s.\n%!"
      (if st.sasarg.rif then "#" else "#")
      str !moves (plur !moves) i (plur i) !rounds (plur !rounds);
    if st.sasarg.rif then 
      print_string "\nq\n#quit\n%!"
    else
      Printf.fprintf log "\n#quit\n"; 
    flush_all();
    i

let () =
  let st = Sasacore.SimuState.make true Sys.argv in
  let log = open_out (st.sasarg.topo ^ ".log") in
  let newdot_fn = (Filename.chop_extension st.sasarg.topo) ^ "_wi.dot" in 
  let newdot = open_out newdot_fn in
  let n = st.sasarg.length in
  try
    match st.sasarg.init_search_max_trials with
    | None ->
      ignore (simuloop stdout n n "" st)
    | Some maxt ->
      let run s =
        moves :=  0;
        rounds :=  0;
        round_mask := [];
        let s = SimuState.update_config s.config s in
        Printf.fprintf log "------------- New simu from %s\n%!"
          (StringOf.env_rif s.config s.network);
        let res =
          try simuloop log n n "" s
          with error ->
            (* We consider that error here mean that a meaningless configuration
               has been tried
            *)
            Printf.fprintf log "==> %s: certainly due to a meaningless configuration\n%!"
              (Printexc.to_string error);
            (-1)
        in
        res
      in
      let st = (WorstInit.fchc log run st maxt) in
      Printf.printf  " (%s)\n%!" (StringOf.env_rif st.config st.network);
      Printf.fprintf newdot "%s\n" (SimuState.to_dot st);
      Printf.printf "%s and %s have been generated\n" (st.sasarg.topo ^ ".log") newdot_fn;
      flush_all();
      close_out newdot;
      close_out log
  with
  | e ->
     Printf.printf "%s%s\n%!" (if st.sasarg.rif then "#" else "") (Printexc.to_string e);
     print_string "\nq\n#quit\n%!";
     flush_all();
     exit 2

